from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec

from ..consts import KEMId
from ..kdf import KDF
from ..kem_key import KEMKey, KEMKeyPair
from ..kem_key_interface import KEMKeyInterface
from ..kem_primitives_interface import KEMPrimitivesInterface
from ..keys.ec_key import ECKey

# from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat


class EC(KEMPrimitivesInterface):
    """
    The KEM (Key Encapsulation Mechanism) context.
    """

    def __init__(self, kem_id: KEMId):
        if kem_id == KEMId.DHKEM_P256_HKDF_SHA256:
            self._crv = ec.SECP256R1()
            self._nsecret = 32
            self._nsk = 32
            self._bitmask = 0xFF
            self._order = 0xFFFFFFFF00000000FFFFFFFFFFFFFFFFBCE6FAADA7179E84F3B9CAC2FC632551
        elif kem_id == KEMId.DHKEM_P384_HKDF_SHA384:
            self._crv = ec.SECP384R1()
            self._nsecret = 48
            self._nsk = 48
            self._bitmask = 0xFF
            self._order = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFC7634D81F4372DDF581A0DB248B0A77AECEC196ACCC52973
        elif kem_id == KEMId.DHKEM_P521_HKDF_SHA512:
            self._crv = ec.SECP521R1()
            self._nsecret = 64
            self._nsk = 66
            self._bitmask = 0x01
            self._order = 0x01FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFA51868783BF2F966B7FCC0148F709A5D03BB5C9B8899C47AEBB6FB71E91386409
        else:
            raise ValueError(f"Invalid kem_id: {kem_id}")

    def generate_key_pair(self) -> KEMKeyPair:
        sk = ec.generate_private_key(self._crv, backend=default_backend())
        pk = sk.public_key()
        return KEMKeyPair(ECKey(sk), ECKey(pk))

    def derive_key_pair(self, ikm: bytes, kdf: KDF) -> KEMKeyPair:
        dkp_prk = kdf.labeled_extract(b"", b"dkp_prk", ikm)

        sk = 0
        counter = 0
        while sk == 0 or sk >= self._order:
            if counter > 255:
                raise ValueError("could not derive keypair")
            raw_key = bytearray(kdf.labeled_expand(dkp_prk, b"candidate", counter.to_bytes(1, "big"), self._nsk))

            raw_key[0] = raw_key[0] & self._bitmask
            sk = int.from_bytes(raw_key, "big")
            counter = counter + 1
        sk_raw = sk.to_bytes(self._nsk, "big", signed=False)

        private_key = self.deserialize_private_key(sk_raw)
        public_key = KEMKey.from_pyca_cryptography_key(private_key._key.public_key())

        return KEMKeyPair(private_key, public_key)

    def deserialize_private_key(self, key: bytes) -> KEMKeyInterface:
        return ECKey.from_private_bytes(self._crv, key)

    def serialize_public_key(self, key: KEMKeyInterface) -> bytes:
        return key.to_public_bytes()

    def deserialize_public_key(self, key: bytes) -> KEMKeyInterface:
        return ECKey.from_public_bytes(self._crv, key)

    def derive_public_key(self, sk: KEMKeyInterface) -> KEMKeyInterface:
        return ECKey(sk.raw.public_key())

    def exchange(self, sk: KEMKeyInterface, pk: KEMKeyInterface) -> bytes:
        return sk.raw.exchange(ec.ECDH(), pk.raw)



from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey

from ..kdf import KDF
from ..kem_key import KEMKey, KEMKeyPair
from ..kem_key_interface import KEMKeyInterface
from ..kem_primitives_interface import KEMPrimitivesInterface
from ..keys.x25519_key import X25519Key

# from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat


class X25519(KEMPrimitivesInterface):
    """ """

    def __init__(self):
        self._nsecret = 32
        self._nsk = 32

    def generate_key_pair(self) -> KEMKeyPair:
        sk = X25519PrivateKey.generate()
        pk = sk.public_key()
        return KEMKeyPair(X25519Key(sk), X25519Key(pk))

    def derive_key_pair(self, ikm: bytes, kdf: KDF) -> KEMKeyPair:
        dkp_prk = kdf.labeled_extract(b"", b"dkp_prk", ikm)
        sk = kdf.labeled_expand(dkp_prk, b"sk", b"", self._nsk)
        private_key = self.deserialize_private_key(sk)
        public_key = KEMKey.from_pyca_cryptography_key(private_key._key.public_key())

        return KEMKeyPair(private_key, public_key)

    def deserialize_private_key(self, key: bytes) -> KEMKeyInterface:
        return X25519Key.from_private_bytes(key)

    def serialize_public_key(self, pk: KEMKeyInterface) -> bytes:
        return pk.to_public_bytes()

    def deserialize_public_key(self, pk: bytes) -> KEMKeyInterface:
        return X25519Key.from_public_bytes(pk)

    def derive_public_key(self, sk: KEMKeyInterface) -> KEMKeyInterface:
        return X25519Key(sk.raw.public_key())

    def exchange(self, sk: KEMKeyInterface, pk: KEMKeyInterface) -> bytes:
        return sk.raw.exchange(pk.raw)



from typing import Any, Dict

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKey,
    EllipticCurvePublicKey,
)
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from ..consts import HPKE_SUPPORTED_JWK_EC_CRVS
from ..kem_key_interface import KEMKeyInterface
from ..utils import base64url_decode


class ECKey(KEMKeyInterface):
    """
    The EC key for KEM.
    """

    def __init__(self, key: Any):
        if isinstance(key, EllipticCurvePrivateKey):
            self._is_public = False
        elif isinstance(key, EllipticCurvePublicKey):
            self._is_public = True
        else:
            raise ValueError("Unsupported key object for EC.")
        super().__init__(key)

    @classmethod
    def from_private_bytes(cls, crv: Any, key: bytes) -> KEMKeyInterface:
        return cls(ec.derive_private_key(int.from_bytes(key, byteorder="big"), crv))

    @classmethod
    def from_public_bytes(cls, crv: Any, key: bytes) -> KEMKeyInterface:
        return cls(ec.EllipticCurvePublicKey.from_encoded_point(crv, key))

    @classmethod
    def from_jwk(cls, jwk: Dict[str, Any]) -> KEMKeyInterface:
        """
        Creates an EC key from JWK (JSON Web Key).
        """
        if "kty" not in jwk or jwk["kty"] != "EC":
            raise ValueError(f"kty is not EC: {jwk['kty']}.")

        if "x" not in jwk:
            raise ValueError("x is not found.")
        if "y" not in jwk:
            raise ValueError("y is not found.")
        x = base64url_decode(jwk["x"])
        y = base64url_decode(jwk["y"])

        if "crv" not in jwk or jwk["crv"] not in HPKE_SUPPORTED_JWK_EC_CRVS:
            raise ValueError(f"Unknown crv: {jwk['crv']}.")
        if jwk["crv"] == "P-256":
            if len(x) == len(y) == 32:
                crv = ec.SECP256R1()
            else:
                raise ValueError("Coords should be 32 bytes for curve P-256")
        elif jwk["crv"] == "P-384":
            if len(x) == len(y) == 48:
                crv = ec.SECP384R1()
            else:
                raise ValueError("Coords should be 48 bytes for curve P-384")
        elif jwk["crv"] == "P-521":
            if len(x) == len(y) == 66:
                crv = ec.SECP521R1()
            else:
                raise ValueError("Coords should be 66 bytes for curve P-521")
        else:
            raise ValueError(f"Invalid curve: {jwk['crv']}.")

        public_numbers = ec.EllipticCurvePublicNumbers(
            x=int.from_bytes(x, byteorder="big"),
            y=int.from_bytes(y, byteorder="big"),
            curve=crv,
        )

        if "d" not in jwk:
            return cls(public_numbers.public_key())

        d = base64url_decode(jwk["d"])
        if len(d) != len(x):
            raise ValueError(f"D should be {len(x)} bytes for curve {jwk['crv']}")

        return cls(ec.EllipticCurvePrivateNumbers(int.from_bytes(d, byteorder="big"), public_numbers).private_key())

    def to_private_bytes(self) -> bytes:
        if self._is_public:
            raise ValueError("The key is public")
        private_value = self._key.private_numbers().private_value  # type: int
        return private_value.to_bytes(self._key.key_size, "big")

    def to_public_bytes(self) -> bytes:
        """
        Serializes the key to a byte string.
        """
        if not self._is_public:
            raise ValueError("The key is private.")
        return self._key.public_bytes(encoding=Encoding.X962, format=PublicFormat.UncompressedPoint)



import json
from typing import Any, Dict, Union

from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKey,
    EllipticCurvePublicKey,
)
from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey, X448PublicKey
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)

from .consts import HPKE_SUPPORTED_JWK_KTYS
from .kem_key_interface import KEMKeyInterface
from .keys.ec_key import ECKey
from .keys.x448_key import X448Key
from .keys.x25519_key import X25519Key


class KEMKey:
    """
    A :class:`KEMKeyInterface <pyhpke.KEMKeyInterface>` Builder.
    """

    @classmethod
    def from_pyca_cryptography_key(cls, k: Any) -> KEMKeyInterface:
        """
        Creates an HPKE key from `pyca/cryptography` key object.
        """
        if isinstance(k, EllipticCurvePrivateKey) or isinstance(k, EllipticCurvePublicKey):
            return ECKey(k)
        elif isinstance(k, X25519PrivateKey) or isinstance(k, X25519PublicKey):
            return X25519Key(k)
        elif isinstance(k, X448PrivateKey) or isinstance(k, X448PublicKey):
            return X448Key(k)
        raise ValueError("Unsupported or unknown key.")

    @classmethod
    def from_jwk(cls, data: Union[bytes, str, Dict[str, Any]]) -> KEMKeyInterface:
        """
        Creates an HPKE key from JWK (JSON Web Key).
        """
        jwk: Dict[str, Any]
        if not isinstance(data, dict):
            jwk = json.loads(data)
        else:
            jwk = data
        if "kty" not in jwk:
            raise ValueError("kty not found.")

        if jwk["kty"] not in HPKE_SUPPORTED_JWK_KTYS:
            raise ValueError(f"Unknown kty: {jwk['kty']}.")
        if jwk["kty"] == "EC":
            return ECKey.from_jwk(jwk)
        if jwk["kty"] == "OKP":
            if "crv" not in jwk:
                raise ValueError("crv not found.")
            if jwk["crv"] == "X25519":
                return X25519Key.from_jwk(jwk)
            if jwk["crv"] == "X448":
                return X448Key.from_jwk(jwk)
            raise ValueError(f"Unsupported or unknown crv: {jwk['crv']}.")
        raise ValueError("Unsupported or unknown key.")

    @classmethod
    def from_pem(cls, data: Union[bytes, str]) -> KEMKeyInterface:
        """
        Creates an HPKE key from PEM-formatted key data.
        """
        if isinstance(data, str):
            data = data.encode("utf-8")
        data_str = data.decode("utf-8")

        k: Any = None
        if "BEGIN PUBLIC" in data_str:
            k = load_pem_public_key(data)
        elif "BEGIN PRIVATE" in data_str:
            k = load_pem_private_key(data, password=None)
        elif "BEGIN EC PRIVATE" in data_str:
            k = load_pem_private_key(data, password=None)
        else:
            raise ValueError("Failed to decode PEM.")
        return cls.from_pyca_cryptography_key(k)


class KEMKeyPair(object):
    def __init__(self, sk: KEMKeyInterface, pk: KEMKeyInterface):
        self._sk = sk
        self._pk = pk
        return

    @property
    def private_key(self) -> KEMKeyInterface:
        return self._sk

    @property
    def public_key(self) -> KEMKeyInterface:
        return self._pk



import json
import os
from typing import Optional

import pytest

from pyhpke import AEADId, CipherSuite, KDFId, KEMId, KEMKeyInterface, KEMKeyPair


class TestWithOfficialTestVectors:
    """
    Tests with official test vectors.
    """

    @pytest.mark.parametrize(
        "v",
        json.load(open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "vectors", "test-vectors.json"))),
    )
    def test_with_official_test_vectors(self, v):
        suite = CipherSuite.new(KEMId(v["kem_id"]), KDFId(v["kdf_id"]), AEADId(v["aead_id"]))

        # deserialize_{private,public}_key
        pke = suite.kem.deserialize_public_key(bytes.fromhex(v["pkEm"]))
        ske = suite.kem.deserialize_private_key(bytes.fromhex(v["skEm"]))
        eks = KEMKeyPair(ske, pke)
        pkr = suite.kem.deserialize_public_key(bytes.fromhex(v["pkRm"]))
        skr = suite.kem.deserialize_private_key(bytes.fromhex(v["skRm"]))
        sks: Optional[KEMKeyInterface] = None
        pks: Optional[KEMKeyInterface] = None
        if "skSm" in v and "pkSm" in v:
            sks = suite.kem.deserialize_private_key(bytes.fromhex(v["skSm"]))
            pks = suite.kem.deserialize_public_key(bytes.fromhex(v["pkSm"]))

        # TODO derive_key_pair
        ikme = bytes.fromhex(v["ikmE"])
        ikme_keypair = suite.kem.derive_key_pair(ikme)
        assert ikme_keypair.private_key.to_private_bytes() == ske.to_private_bytes()
        assert ikme_keypair.public_key.to_public_bytes() == pke.to_public_bytes()

        ikmr = bytes.fromhex(v["ikmR"])
        ikmr_keypair = suite.kem.derive_key_pair(ikmr)
        assert ikmr_keypair.private_key.to_private_bytes() == skr.to_private_bytes()
        assert ikmr_keypair.public_key.to_public_bytes() == pkr.to_public_bytes()

        # create_{sender,recipient}_context
        info = bytes.fromhex(v["info"])
        psk = bytes.fromhex(v["psk"]) if "psk" in v else b""
        psk_id = bytes.fromhex(v["psk_id"]) if "psk_id" in v else b""
        enc, sender = suite.create_sender_context(pkr, info, sks, psk, psk_id, eks)
        assert enc == bytes.fromhex(v["enc"])
        recipient = suite.create_recipient_context(enc, skr, info, pks, psk, psk_id)
        if v["aead_id"] != 0xFFFF:
            for ve in v["encryptions"]:
                pt = bytes.fromhex(ve["pt"])
                aad = bytes.fromhex(ve["aad"])
                ct = bytes.fromhex(ve["ct"])
                sealed = sender.seal(pt, aad)
                opened = recipient.open(sealed, aad)
                assert pt == opened
                assert ct == sealed
            return
        for ve in v["exports"]:
            ec = bytes.fromhex(ve["exporter_context"]) if len(ve["exporter_context"]) > 0 else b""
            ev = bytes.fromhex(ve["exported_value"])
            exported = sender.export(ec, ve["L"])
            assert ev == exported
            exported = recipient.export(ec, ve["L"])
            assert ev == exported



from typing import Any, Dict

from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey, X448PublicKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from ..kem_key_interface import KEMKeyInterface
from ..utils import base64url_decode


class X448Key(KEMKeyInterface):
    """
    The X448 key for KEM.
    """

    def __init__(self, key: Any):
        if isinstance(key, X448PrivateKey):
            self._is_public = False
        elif isinstance(key, X448PublicKey):
            self._is_public = True
        else:
            raise ValueError("Unsupported key object for EC.")
        super().__init__(key)

    @classmethod
    def from_private_bytes(cls, key: bytes) -> KEMKeyInterface:
        return cls(X448PrivateKey.from_private_bytes(key))

    @classmethod
    def from_public_bytes(cls, key: bytes) -> KEMKeyInterface:
        return cls(X448PublicKey.from_public_bytes(key))

    @classmethod
    def from_jwk(cls, jwk: Dict[str, Any]) -> KEMKeyInterface:
        """
        Creates an EC key from JWK (JSON Web Key).
        """
        if "kty" not in jwk or jwk["kty"] != "OKP":
            raise ValueError(f"kty is not OKP: {jwk['kty']}.")

        if "x" not in jwk:
            raise ValueError("x is not found.")
        x = base64url_decode(jwk["x"])
        if len(x) != 56:
            raise ValueError("Coords should be 56 bytes for X448.")

        if "crv" not in jwk or jwk["crv"] != "X448":
            raise ValueError(f"Unknown crv: {jwk['crv']}.")

        if "d" not in jwk:
            return cls(X448PublicKey.from_public_bytes(x))

        d = base64url_decode(jwk["d"])
        if len(d) != len(x):
            raise ValueError(f"D should be {len(x)} bytes for {jwk['crv']}")
        return cls(X448PrivateKey.from_private_bytes(d))

    def to_private_bytes(self) -> bytes:
        if self._is_public:
            raise ValueError("The key is public")
        return self._key.private_bytes_raw()

    def to_public_bytes(self) -> bytes:
        """
        Serializes the key to a byte string.
        """
        if not self._is_public:
            raise ValueError("The key is private.")
        return self._key.public_bytes(Encoding.Raw, PublicFormat.Raw)



import struct
from typing import Any

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac

from .consts import HPKE_VERSION, KDFId
from .kdf_interface import KDFInterface


class KDF(KDFInterface):
    """
    A KDF(Key Derivation Function) context.
    """

    def __init__(self, kdf_id: KDFId, suite_id: bytes):
        h: Any = None

        if kdf_id == KDFId.HKDF_SHA256:
            h = hashes.SHA256()
        elif kdf_id == KDFId.HKDF_SHA384:
            h = hashes.SHA384()
        elif kdf_id == KDFId.HKDF_SHA512:
            h = hashes.SHA512()
        else:
            raise ValueError(f"Unsupported or unknown KDF id: {kdf_id}.")

        self._id = kdf_id
        self._hash = h
        self._suite_id = suite_id
        return

    @property
    def id(self) -> KDFId:
        """
        The KDF identifier.
        """
        return self._id

    @property
    def digest_size(self) -> int:
        return self._hash.digest_size

    def extract(self, salt: bytes, ikm: bytes) -> bytes:
        ctx = hmac.HMAC(salt, self._hash, backend=default_backend())
        ctx.update(ikm)
        return ctx.finalize()

    def expand(self, prk: bytes, info: bytes, length: int) -> bytes:
        assert length <= 255 * self._hash.digest_size

        t_n_minus_1 = b""
        n = 1
        data = b""

        while len(data) < length:
            ctx = hmac.HMAC(prk, self._hash, backend=default_backend())
            ctx.update(t_n_minus_1 + info + n.to_bytes(1, byteorder="big"))
            t_n_minus_1 = ctx.finalize()
            data += t_n_minus_1
            n += 1
        return data[:length]

    def extract_and_expand(self, salt: bytes, ikm: bytes, info: bytes, length: int) -> bytes:
        return self.expand(self.extract(salt, ikm), info, length)

    def labeled_extract(self, salt: bytes, label: bytes, ikm: bytes) -> bytes:
        labeled_ikm = HPKE_VERSION + self._suite_id + label + ikm
        return self.extract(salt, labeled_ikm)

    def labeled_expand(self, prk: bytes, label: bytes, info: bytes, length: int) -> bytes:
        labeled_info = struct.pack(">H", length) + HPKE_VERSION + self._suite_id + label + info
        return self.expand(prk, labeled_info, length)



from .context_interface import ContextInterface
from .exceptions import NotSupportedError
from .kdf import KDF


class ExporterContext(ContextInterface):
    def __init__(self, kdf: KDF, exporter_secret: bytes):
        self._kdf = kdf
        self._exporter_secret = exporter_secret
        return

    def seal(self, pt: bytes, aad: bytes = b"") -> bytes:
        raise NotSupportedError("Not available on export-only mode.")

    def open(self, ct: bytes, aad: bytes = b"") -> bytes:
        raise NotSupportedError("Not available on export-only mode.")

    def export(self, exporter_context: bytes, length: int) -> bytes:
        return self._kdf.labeled_expand(self._exporter_secret, b"sec", exporter_context, length)



import struct
from typing import Optional, Tuple

from .aead import AEAD, AEADParams
from .aead_interface import AEADInterface
from .consts import AEADId, KDFId, KEMId, Mode
from .context_interface import ContextInterface
from .exporter_context import ExporterContext
from .kdf import KDF
from .kdf_interface import KDFInterface
from .kem import KEM
from .kem_interface import KEMInterface
from .kem_key import KEMKeyPair
from .kem_key_interface import KEMKeyInterface
from .recipient_context import RecipientContext
from .sender_context import SenderContext


class CipherSuite(object):
    """
    An HPKE cipher suite which consists of KEM, KDF and AEAD.
    """

    def __init__(self, kem: KEMInterface, kdf: KDFInterface, aead: AEADInterface):
        self._kem = kem
        self._kdf = kdf
        self._aead = aead
        return

    @classmethod
    def new(cls, kem_id: KEMId, kdf_id: KDFId, aead_id: AEADId):
        """
        Constructor of HPKE cipher suite.

        Args:
            kem_id (KEMId): A KEM (Key Encapsulation Mechanism) identifier.
            kdf_id (KDFId): A KDF (Key Derivation Function) identifier.
            aead_id (AEADId): An AEAD (Authenticated Encryption with Additional Data) identifier.
        Returns:
            bytes: A CipherSuite object.
        """
        suite_id = b"HPKE" + struct.pack(">HHH", kem_id.value, kdf_id.value, aead_id.value)
        kem = KEM(kem_id)
        kdf = KDF(kdf_id, suite_id)
        aead = AEAD(aead_id)
        return cls(kem, kdf, aead)

    @property
    def kem(self) -> KEMInterface:
        """
        The KEM context in the cipher suite.
        """
        return self._kem

    @property
    def kdf(self) -> KDFInterface:
        """
        The KDF context in the cipher suite.
        """
        return self._kdf

    @property
    def aead(self) -> AEADInterface:
        """
        The AEAD context in the cipher suite.
        """
        return self._aead

    def create_sender_context(
        self,
        pkr: KEMKeyInterface,
        info: bytes = b"",
        sks: Optional[KEMKeyInterface] = None,
        psk: bytes = b"",
        psk_id: bytes = b"",
        eks: Optional[KEMKeyPair] = None,
    ) -> Tuple[bytes, ContextInterface]:
        """
        Creates a sender context.
        """
        mode: Mode = Mode.BASE
        if psk == b"":
            mode = Mode.BASE if sks is None else Mode.AUTH
        else:
            mode = Mode.PSK if sks is None else Mode.AUTH_PSK

        shared_secret, enc = self._kem.encap(pkr, sks, eks)
        return enc, self._key_schedule_s(mode, shared_secret, info, psk, psk_id)

    def create_recipient_context(
        self,
        enc: bytes,
        skr: KEMKeyInterface,
        info: bytes = b"",
        pks: Optional[KEMKeyInterface] = None,
        psk: bytes = b"",
        psk_id: bytes = b"",
    ) -> ContextInterface:
        """
        Creates a recipient context.
        """
        mode: Mode = Mode.BASE
        if psk == b"":
            mode = Mode.BASE if pks is None else Mode.AUTH
        else:
            mode = Mode.PSK if pks is None else Mode.AUTH_PSK

        shared_secret = self._kem.decap(enc, skr, pks)
        return self._key_schedule_r(mode, shared_secret, info, psk, psk_id)

    def seal(
        self,
        pkr: KEMKeyInterface,
        pt: bytes,
        info: bytes = b"",
        aad: bytes = b"",
        psk: bytes = b"",
        psk_id: bytes = b"",
        sks: Optional[KEMKeyInterface] = None,
    ) -> Tuple[bytes, bytes]:
        """
        Does a single-shot encryption.
        """
        raise NotImplementedError()

    def open(
        self,
        enc: bytes,
        skr: KEMKeyInterface,
        ct: bytes,
        info: bytes = b"",
        aad: bytes = b"",
        psk: bytes = b"",
        psk_id: bytes = b"",
        pks: Optional[KEMKeyInterface] = None,
    ) -> bytes:
        """
        Does a single-shot decryption.
        """
        raise NotImplementedError()

    def _key_schedule(
        self,
        mode: Mode,
        shared_secret: bytes,
        info: bytes,
        psk: bytes,
        psk_id: bytes,
    ) -> Tuple[KDF, AEADParams]:
        suite_id = b"HPKE" + struct.pack(">HHH", self._kem.id.value, self._kdf.id.value, self._aead.id.value)
        kdf = KDF(self._kdf.id, suite_id)

        psk_id_hash = kdf.labeled_extract(b"", b"psk_id_hash", psk_id)
        info_hash = kdf.labeled_extract(b"", b"info_hash", info)
        key_schedule_context = bytes([mode.value]) + psk_id_hash + info_hash

        secret = kdf.labeled_extract(shared_secret, b"secret", psk)

        key = kdf.labeled_expand(secret, b"key", key_schedule_context, self._aead.key_size)
        base_nonce = kdf.labeled_expand(secret, b"base_nonce", key_schedule_context, self._aead.nonce_size)
        exporter_secret = kdf.labeled_expand(secret, b"exp", key_schedule_context, kdf.digest_size)
        return kdf, AEADParams(self._aead, key, base_nonce, 0, exporter_secret)

    def _key_schedule_s(
        self,
        mode: Mode,
        shared_secret: bytes,
        info: bytes,
        psk: bytes,
        psk_id: bytes,
    ) -> ContextInterface:
        kdf, params = self._key_schedule(mode, shared_secret, info, psk, psk_id)
        if params.key == b"":
            return ExporterContext(kdf, params.exporter_secret)
        return SenderContext(kdf, params)

    def _key_schedule_r(
        self,
        mode: Mode,
        shared_secret: bytes,
        info: bytes,
        psk: bytes,
        psk_id: bytes,
    ) -> ContextInterface:
        kdf, params = self._key_schedule(mode, shared_secret, info, psk, psk_id)
        if params.key == b"":
            return ExporterContext(kdf, params.exporter_secret)
        return RecipientContext(kdf, params)



class ContextInterface(object):
    def seal(self, pt: bytes, aad: bytes = b"") -> bytes:
        raise NotImplementedError()

    def open(self, ct: bytes, aad: bytes = b"") -> bytes:
        raise NotImplementedError()

    def export(self, exporter_context: bytes, length: int) -> bytes:
        raise NotImplementedError()



class AEADKeyInterface:
    """
    The AEAD key interface.
    """

    def seal(self, pt: bytes, nonce: bytes, aad: bytes = b"") -> bytes:
        """
        Encrypts the specified message.

        Args:
            pt (bytes): A plain text to be encrypted.
            nonce (bytes): A nonce for encryption.
            aad (bytes): Additional authenticated data.
        Returns:
            bytes: The encrypted data.
        Raises:
            ValueError: Invalid arguments.
            SealError: Failed to encrypt the plain text.
        """
        raise NotImplementedError()

    def open(self, ct: bytes, nonce: bytes, aad: bytes = b"") -> bytes:
        """
        Decrypts the specified message.

        Args:
            ct (bytes): A cipher text to be decrypted.
            nonce (bytes): A nonce for encryption.
            aad (bytes): Additional authenticated data.
        Returns:
            bytes: The decrypted data.
        Raises:
            ValueError: Invalid arguments.
            OpenError: Failed to decrypt the cipher text.
        """
        raise NotImplementedError()



import struct
from typing import Any, Optional, Tuple

from .consts import KDFId, KEMId
from .kdf import KDF
from .kem_interface import KEMInterface
from .kem_key import KEMKeyPair
from .kem_key_interface import KEMKeyInterface
from .kem_primitives.ec import EC
from .kem_primitives.x448 import X448
from .kem_primitives.x25519 import X25519


class KEM(KEMInterface):
    """
    The KEM (Key Encapsulation Mechanism) interface.
    """

    def __init__(self, kem_id: KEMId):
        self._id = kem_id
        self._prim: Any
        if kem_id == KEMId.DHKEM_P256_HKDF_SHA256:
            self._nsecret = 32
            kdf_id = KDFId.HKDF_SHA256
            self._prim = EC(kem_id)
        elif kem_id == KEMId.DHKEM_P384_HKDF_SHA384:
            self._nsecret = 48
            kdf_id = KDFId.HKDF_SHA384
            self._prim = EC(kem_id)
        elif kem_id == KEMId.DHKEM_P521_HKDF_SHA512:
            kdf_id = KDFId.HKDF_SHA512
            self._nsecret = 64
            self._prim = EC(kem_id)
        elif kem_id == KEMId.DHKEM_X25519_HKDF_SHA256:
            kdf_id = KDFId.HKDF_SHA256
            self._nsecret = 32
            self._prim = X25519()
        elif kem_id == KEMId.DHKEM_X448_HKDF_SHA512:
            kdf_id = KDFId.HKDF_SHA512
            self._nsecret = 64
            self._prim = X448()
        else:
            raise ValueError("The specified kem is not supported.")
        suite_id = b"KEM" + struct.pack(">H", kem_id.value)
        self._kdf = KDF(kdf_id, suite_id)
        return

    @property
    def id(self) -> KEMId:
        """
        The AEAD identifier.
        """
        return self._id

    def derive_key_pair(self, ikm: bytes) -> KEMKeyPair:
        return self._prim.derive_key_pair(ikm, self._kdf)

    def extract_and_expand(self, dh: bytes, kem_context: bytes, length: int) -> bytes:
        eae_prk = self._kdf.labeled_extract(b"", b"eae_prk", dh)
        shared_secret = self._kdf.labeled_expand(eae_prk, b"shared_secret", kem_context, length)
        return shared_secret

    def deserialize_private_key(self, key: bytes) -> KEMKeyInterface:
        return self._prim.deserialize_private_key(key)

    def deserialize_public_key(self, key: bytes) -> KEMKeyInterface:
        return self._prim.deserialize_public_key(key)

    def encap(
        self,
        pkr: KEMKeyInterface,
        sks: Optional[KEMKeyInterface] = None,
        eks: Optional[KEMKeyPair] = None,
    ) -> Tuple[bytes, bytes]:
        """ """
        if eks is None:
            ek = self._prim.generate_key_pair()
        else:
            # For testing purpose only
            ek = eks
        enc = self._prim.serialize_public_key(ek.public_key)

        if sks is None:
            dh = self._prim.exchange(ek.private_key, pkr)
            kem_context = enc + self._prim.serialize_public_key(pkr)
        else:
            dh1 = self._prim.exchange(ek.private_key, pkr)
            dh2 = self._prim.exchange(sks, pkr)
            dh = dh1 + dh2
            pks = self._prim.derive_public_key(sks)
            kem_context = enc + self._prim.serialize_public_key(pkr) + self._prim.serialize_public_key(pks)

        shared_secret = self.extract_and_expand(dh, kem_context, self._nsecret)
        return shared_secret, enc

    def decap(self, enc: bytes, skr: KEMKeyInterface, pks: Optional[KEMKeyInterface] = None) -> bytes:
        """ """
        pke = self._prim.deserialize_public_key(enc)
        pkr = self._prim.derive_public_key(skr)
        if pks is None:
            dh = self._prim.exchange(skr, pke)
            kem_context = enc + self._prim.serialize_public_key(pkr)
        else:
            dh1 = self._prim.exchange(skr, pke)
            dh2 = self._prim.exchange(skr, pks)
            dh = dh1 + dh2
            kem_context = enc + self._prim.serialize_public_key(pkr) + self._prim.serialize_public_key(pks)

        shared_secret = self.extract_and_expand(dh, kem_context, self._nsecret)
        return shared_secret



from .aead import AEADParams
from .exporter_context import ExporterContext
from .kdf import KDF
from .utils import xor_bytes


class EncryptionContext(ExporterContext):
    def __init__(self, kdf: KDF, aead_params: AEADParams):
        self._key = aead_params.ctx.import_key(aead_params.key)
        self._seq = aead_params.seq
        self._nonce = aead_params.base_nonce
        self._nn = aead_params.ctx.nonce_size
        super().__init__(kdf, aead_params.exporter_secret)
        return

    def _next_nonce(self):
        nonce = xor_bytes(self._nonce, self._seq.to_bytes(self._nn, byteorder="big"))
        self._seq += 1
        return nonce



from .encryption_context import EncryptionContext
from .exceptions import NotSupportedError, SealError


class SenderContext(EncryptionContext):
    def seal(self, pt: bytes, aad: bytes = b"") -> bytes:
        try:
            return self._key.seal(pt, self._next_nonce(), aad)
        except Exception as err:
            raise SealError("Failed to seal.") from err

    def open(self, ct: bytes, aad: bytes = b"") -> bytes:
        raise NotSupportedError("Not available for sender.")
